# SU extension dl_camera_spiral
#
# Copyright © 2016 De Luminae
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

require 'sketchup'

module DL::CameraSpiral
    if Sketchup.version.to_f < 16
        UI.messagebox("You must have Sketchup 2016 to run the plugin "\
                      "DL Camera Spiral.")
        OC_MENU_DONE = true  # hack to no display menu if bad version
    end
    # put link in menu
    if !defined? OC_MENU_DONE
        folder_cmd = UI::Command.new('Select images folder') {
            choose_folder
        }
        conf_cmd = UI::Command.new('Edit config') {
            set_config
        }
        set_orig_cmd = UI::Command.new('Set center position') {
            set_orig
        }
        set_start_cmd = UI::Command.new('Set start view') {
            set_start
        }
        set_end_cmd = UI::Command.new('Set end view') {
            set_end
        }
        show_cmd = UI::Command.new("Make show") {
            make_show
        }
        stop_cmd = UI::Command.new("Stop") {
            stop
        }
        images_cmd = UI::Command.new("Make images") {
            make_images
        }
        help_cmd = UI::Command.new("Help") {
            help_screen
        }
        about_cmd = UI::Command.new("About") {
            about_screen
        }
        menu = UI.menu("Plugins").add_submenu("DL Camera Spiral")
        menu.add_item(folder_cmd)
        menu.add_item(conf_cmd)
        menu.add_separator
        menu.add_item(set_orig_cmd)
        menu.add_item(set_start_cmd)
        menu.add_item(set_end_cmd)
        menu.add_separator
        menu.add_item(show_cmd)
        menu.add_item(images_cmd)
        menu.add_item(stop_cmd)
        menu.add_separator
        menu.add_item(help_cmd)
        menu.add_item(about_cmd)
        OC_MENU_DONE = true
    end

    @folder = nil
    @names_config = ['Center',
                     'Shifting',
                     'Duration (s)',
                     'Loop',
                     'Rotations',
                     'FPS',
                     'Images prefix',
                     'Width',
                     'Height',
                     'Antialias',
                     'Image Type',
                     'Transparency',
                     'Image exists']
    @default_config = ['progressive',
                       'altitude',
                       10,
                       'false',
                       0.0,
                       25,
                       'img',
                       1920,
                       1080,
                       'true',
                       '.jpg',
                       'false',
                       'skip frame']
    @config = nil
    @do_images = false
    @running = false
    @start_pos = nil
    @end_pos = nil
    @html_initialized = false
    @is_osx = nil
    @logo=nil
    @help_html = nil
    @about_html = nil

    class << self
        attr_accessor :folder
        attr_accessor :config
        attr_accessor :do_images
        attr_accessor :running
        attr_accessor :start_pos
        attr_accessor :end_pos
    end


    def self.initialize_html
        @html_initialized && return
        @is_osx = Sketchup.platform == :platform_osx
        @logo = File.join(File.dirname(__FILE__), 'deluminae.png')

        @help_html = %Q(<html>
<head>
<title>Camera Spiral</title>
</head>
<body style="background-color:#ffffff;font:message-box;">
  <img alt="logo" src="#{@logo}" style="display:block;margin:0 auto;">
  <div style="padding:4px 4px 8px 8px;">
    <p>
    <strong>Camera Spiral</strong> is a basic utility to rotate camera around
    some center origin, radius and distance evolving between two positions, to
    make a show or generate images.
    </p>
    <p>Menu:
    <ul>
        <li>Select images folder: select the directory were computed images are
        saved.</li>
        <li>Edit config: choose various configuration options, see below.</li>
        <li>Set center position: define center point of the rotation. If not
        defined, the axes origin of the view is used instead.</li>
        <li>Set start view: choose current view as starting position.</li>
        <li>Set end view: choose current view as ending position.</li>
        <li>Make show: only display the rotating show, no images created.</li>
        <li>Make images: display the rotating show and create images.</li>
        <li>Stop: stop show and images computation.</li>
        <li>Help: show this message.</li>
        <li>About: about this utility.</li>
    </ul>
    <br>
    Configuration:
    <ul>
        <li>Center: center the camera view on either center origin, start
        position or progressive move from start to end position.</li>
        <li>Shifting: orient the camera view either by modifying altitude
        (recommanded) or angle of view.</li>
        <li>Duration: duration of rotation in seconds.</li>
        <li>Loop: for show only, keep rotating for ever.</li>
        <li>Rotations: If 0, use the angle between start and end position.
        If non zero, it's the number of rotations, 1.0 for one full circle,
        negative to reverse direction, 2.5 for 2 turns and half.</li>
        <li>FPS: for images generation, number of frames per second.</li>
        <li>Images prefix: for images generation, prefix of the generated images
        files.</li>
        <li>Width: for images generation, width in pixels of images.</li>
        <li>Height: for images generation, height in pixels of images.</li>
        <li>Antialias: for images generation, use antialias.</li>
        <li>Image type: for images generation, choose either JPG or PNG
        format.</li>
        <li>Transparency: for images generation, set background to
        transparent.</li>
        <li>Image exists: for images generation, if target file already exists,
        either ower write with new one or skip this frame, permitting to
        interrupt/continue some long rendering.</li>
    </ul>
    </p>
  </div>
</body>
</html>)

        @about_html = %Q(<html>
<head>
<title>About Camera Spiral</title>
</head>
<body style="background-color:#ffffff;font:message-box;">
  <img alt="logo" src="#{@logo}" style="display:block;margin:0 auto;">
  <div style="padding:4px 4px 0px 8px;">
    <p>
    <strong>Camera Spiral</strong> is a basic utility to rotate camera around
    some center origin, radius and distance evolving between two positions, to
    make a show or generate images.
    </p>
    <p style="padding-left:20px;">
    Version: 1.5<br>
    Author: <a href="http://www.deluminaelab.com" target="blank">De Luminae</a> (MIT license)<br>
    <br>
    Find us on Facebook: <a href="https://www.facebook.com/deluminae" target="blank">deluminae</a><br>
    </p>
  </div>
</body>
</html>)
        @html_initialized = true
    end

    def self.about_screen
        initialize_html
        if @is_osx
            dlg = UI::WebDialog.new("About Camera Spiral", true, nil, 300, 270,
                                    100, 100, true)
        else
            dlg = UI::WebDialog.new("About Camera Spiral", false, nil, 300, 310,
                                    100, 100, true)
        end
        dlg.set_html(@about_html)
        if @is_osx
            dlg.show_modal
        else
            dlg.show
            dlg.bring_to_front
        end
    end

    def self.help_screen
        initialize_html
        dlg = UI::WebDialog.new("Camera Spiral", true, nil, 450, 550, 150, 150,
                                true)
        dlg.set_html(@help_html)
        if @is_osx
            dlg.show_modal
        else
            dlg.show
            dlg.bring_to_front
        end
    end

    def self.choose_folder
        if @folder
            start_folder = @folder
        else
            start_folder = File.expand_path('~')
        end
        @folder = UI.select_directory(title: 'Select image folder',
                                      directory: start_folder,
                                      select_multiple: false)
    end

    def self.set_config
        if @config
            default = @config
        else
            default = @default_config
        end

        @config = UI.inputbox(
            @names_config,
            default,
            ['origin|start position|progressive',
             'altitude|angle',
             nil,
             'true|false',
             nil,
             nil,
             nil,
             nil,
             nil,
             'true|false',
             '.jpg|.png',
             'true|false',
             'skip frame|overwrite'],
             'Camera Spiral configuration')
    end

    def self.set_pos
        model = Sketchup.active_model
        view = model.active_view
        return {:e => view.camera.eye.clone, :t => view.camera.target.clone}
    end

    def self.set_start
        @start_pos = set_pos
    end

    def self.set_end
        @end_pos = set_pos
    end

    def self.set_orig
        Sketchup.active_model.select_tool OrigTool.new
    end

    def self.check_ready
        if !@start_pos
            set_start
        end
        if !@end_pos
            @end_pos = @start_pos
        end
        if !@config
            set_config
        end
        if !@folder and @do_images
            choose_folder
        end
    end

    def self.make_show
        @do_images = false
        check_ready
        do_orbit
    end

    def self.stop
        @running = false
    end

    def self.make_images
        @do_images = true
        check_ready
        do_orbit
    end

    def self.do_orbit
        @running = true
        Sketchup.active_model.active_view.animation = Orbiter.new
    end


    class OrigTool
        @@op = nil

        def OrigTool.get_op
            return @@op
        end

        def initialize
            @op = nil
        end

        def activate
            @op = Sketchup::InputPoint.new
            @p = Sketchup::InputPoint.new
            @chosen = false
            self.reset(nil)
        end

        def reset(view)
            @chosen = false
            @op.clear
            if view
                view.tooltip = nil
                view.invalidate
            end
        end

        def deactivate(view)
            view.invalidate
        end

        def suspend(view)
            view.invalidate
        end

        def resume(view)
            view.invalidate
        end

        def onMouseMove(flags, x, y, view)
            if !@chosen
                @p.pick view, x, y
                if (@p != @op)
                    (@op.display? or @p.display?) && view.invalidate
                    @op.copy! @p
                    view.tooltip = @op.tooltip
                end
            end
        end

        def onLButtonDown(flags, x, y, view)
            if !@chosen
                @op.pick view, x, y
                if (@op.valid?)
                    @chosen = true
                    @@op = @op.position
                end
            end
            view.lock_inference
        end

        def onLButtonUp(flags, x, y, view)
            deactivate(view)
        end

        def onKeyDown(key, repeat, flags, view)
            if (key == CONSTRAIN_MODIFIER_KEY && repeat == 1)
                @shift_down_time = Time.now
                if view.inference_locked?
                    view.lock_inference
                elsif (!@chosen && @op.valid?)
                    view.lock_inference @op
                end
            end
        end

        def onKeyUp(key, repeat, flags, view)
            if (key == CONSTRAIN_MODIFIER_KEY &&
                view.inference_locked? &&
                (Time.now - @shift_down_time) > 0.5)
                view.lock_inference
            end
        end

        def draw(view)
            if @op.valid?
                if @op.display?
                    @op.draw(view)
                end
            end
        end

        def onCancel(flag, view)
            self.reset(view)
        end
    end # class OrigTool

    class Orbiter
        def initialize
            conf = DL::CameraSpiral.config
            @center_origin = conf[0] == 'origin'
            @progressive = conf[0] == 'progressive'
            @altitude = conf[1] == 'altitude'
            @duration = conf[2]
            @loop = conf[3] == 'true'
            @rotations = conf[4]
            @fps = conf[5]
            @prefix = conf[6]
            @width = conf[7]
            @height = conf[8]
            @antialias = conf[9] == 'true'
            @imgtype = conf[10]
            @transparent = conf[11] == 'true'
            @overwrite = conf[12] == 'overwrite'
            @start_pos = DL::CameraSpiral.start_pos
            @end_pos = DL::CameraSpiral.end_pos
            @zero = Geom::Point3d.new 0, 0, 0
            if DL::CameraSpiral.do_images
                @loop = false
            end
            model = Sketchup.active_model
            # origin
            @o = OrigTool.get_op
            !@o && @o = model.axes.origin.clone
            # eye position
            @e0 = @start_pos[:e] - @o
            @e1 = @end_pos[:e] - @o
            # target position
            t = @start_pos[:t] - @o
            @t0 = projection_target_on_origin_plane(@e0, t)
            t = @end_pos[:t] - @o
            @t1 = projection_target_on_origin_plane(@e1, t)

            @steps = @duration * @fps - 1  # first image is initial scene

            @delta_radius = (@e1.length - @e0.length) / @steps
            vertical = Geom::Vector3d.new 0, 0, 1
            if @rotations == 0
                # use actual e1 e0 angle
                d0 = Math::sqrt(@e0.x**2 + @e0.y**2)
                a0 = Math::acos(@e0.x/d0)
                if @e0.y > 0  # invert rotation
                    a0 = 2 * Math::PI - a0
                end
                rot_a0 = Geom::Transformation.rotation @zero, vertical, a0
                e2 = @e1.transform rot_a0
                d2 = Math::sqrt(e2.x**2 + e2.y**2)
                angle = Math::acos(e2.x/d2)
                if e2.y < 0
                    angle = 2 * Math::PI - angle
                end
                if angle > Math::PI
                    angle = angle - 2 * Math::PI
                end
                actual_rotations = angle / 2 / Math::PI
            else
                actual_rotations = @rotations
            end
            d_angle = actual_rotations * 2 * Math::PI / @steps
            @tr_rot = Geom::Transformation.rotation @zero, vertical, d_angle

            if @altitude
                # variation by altitude
                @e_dz = (@e1.z - @e0.z) / @steps
                @e_beta = nil
            else
                # variation vertical view angle
                v = Geom::Vector3d.new @e0.x, @e0.y, 0
                b0 = @e0.angle_between v
                v = Geom::Vector3d.new @e1.x, @e1.y, 0
                b1 = @e1.angle_between v
                @e_beta = (b1 - b0) / @steps
                @e_dz = nil
            end

            init_counters
        end

        def init_counters
            # moving eye and targets
            @e = @e0.clone
            @mt0 = @t0.clone
            @mt1 = @t1.clone
            @count = 0
            @skipped = 0
            @time0 = Time.now
        end

        def projection_target_on_origin_plane(e, t)
            e_vector = Geom::Vector3d.new e.x, e.y, e.z
            plane = [@zero, e_vector]
            line = [Geom::Point3d.new(e.x, e.y, e.z),
                    Geom::Point3d.new(t.x, t.y, t.z)]
            proj_target = Geom.intersect_line_plane(line, plane)
            return proj_target - @zero  # convert to vector
        end

        def rotate_beta(vector, beta)
            ortho = Geom::Vector3d.new(vector.y, -vector.x, 0)
            rotation = Geom::Transformation.rotation @zero, ortho, beta
            vector.transform! rotation
        end

        def compute(view)
            if @count > 0  # because first image is initial scene
                @e.transform! @tr_rot
                @e.length = @e.length + @delta_radius
                if @altitude
                    z = @e0.z + @e_dz * @count
                else
                    rotate_beta(@e, @e_beta)
                    z = @e.z
                end
            else
                z = @e.z
            end
            eye = Geom::Point3d.new(@e.x, @e.y, z) + [@o.x, @o.y, @o.z]

            if @center_origin
                target = @o
            else
                if @count > 0
                    @mt0.transform! @tr_rot
                    @mt1.transform! @tr_rot
                end
                r = 1.0 * @count / @steps
                if @progressive
                    x = (1 - r) * @mt0.x + r * @mt1.x
                    y = (1 - r) * @mt0.y + r * @mt1.y
                    z = (1 - r) * @mt0.z + r * @mt1.z
                else
                    x = @mt0.x
                    y = @mt0.y
                    z = (1 - r) * @mt0.z + r * @mt1.z
                end
                target = Geom::Point3d.new(x, y, z) + [@o.x, @o.y, @o.z]
            end
            view.camera.set(eye, target, [0, 0, 1])
        end

        def dump_image(view)
            filename = File.join(DL::CameraSpiral.folder,
                                 @prefix + "%03d" % @count + @imgtype)
            if !@overwrite
                if File.file?(filename)
                    @skipped += 1
                    return
                end
            end
            keys_j = {
                :filename => filename,
                :width => @width,
                :height => @height,
                :antialias => @antialias,
                :compression => 1.0,
                :transparent => @transparent
                }
            view.write_image keys_j
        end

        def nextFrame(view)
            compute(view)
            view.show_frame
            DL::CameraSpiral.do_images && dump_image(view)
            @count += 1
            if DL::CameraSpiral.do_images
                actual = @count - @skipped
                pc = "Camera spiral %.1f%" % (100.0 *  @count / (@steps + 1))
                if actual > 1
                    eta = (@steps + 1 - @count) * (Time.now - @time0) / actual
                    h = eta.div 3600
                    eta = eta - 3600 * h
                    m = eta.div 60
                    s = eta.to_i % 60
                    if h > 0
                        etas = "#{h}h #{m}min #{s}s"
                    else
                        etas = "#{m}min #{s}s"
                    end
                else
                    etas = ""
                end
                Sketchup.set_status_text(("#{pc}  ETA: #{etas}" ), SB_PROMPT)
            else
                advance = @count - (Time.now - @time0) * @fps
                if advance > 0
                    sleep(advance / @fps)
                end
            end
            continue = DL::CameraSpiral.running && (@loop || (@count <= @steps))
            # loop restarting
            if @loop && (@count > @steps)
                init_counters
            end
            if continue
                return true
            else
                DL::CameraSpiral.running = false
                if DL::CameraSpiral.do_images
                    Sketchup.set_status_text(("#{pc}  stopped." ), SB_PROMPT)
                end
                return false
            end
        end
    end
end
